@@ -21,6 +21,7 @@
 ;;}}}
 ;;{{{ Load some required packages
 (require 'mailcrypt)
+(require 'mail-utils)
 
 (eval-when-compile
   ;; RMAIL
@@ -52,9 +53,7 @@
   ;; Takes a comma separated string of recipients to encrypt for and,
   ;; assuming they were possibly extracted from the headers of a reply,
   ;; returns a list of the address components.
-  (mapcar (function
-	   (lambda (x)
-	     (car (cdr (mail-extract-address-components x)))))
+  (mapcar 'mc-strip-address
 	  (mc-split "\\([ \t\n]*,[ \t\n]*\\)+" str)))
 
 (defun mc-find-headers-end ()
@@ -99,6 +98,10 @@ With \\[universal-argument] \\[universal-argument], prompt for encryption scheme
 (defun mc-encrypt-generic (&optional recipients scheme start end from sign)
   "*Generic function to encrypt a region of data."
   (save-excursion
+    (or start (setq start (point-min-marker)))
+    (or (markerp start) (setq start (copy-marker start)))
+    (or end (setq end (point-max-marker)))
+    (or (markerp end) (setq end (copy-marker end)))
     (run-hooks 'mc-pre-encryption-hook)
     (cond ((stringp recipients)
 	   (setq recipients
@@ -108,8 +111,6 @@ With \\[universal-argument] \\[universal-argument], prompt for encryption scheme
 		 (mc-cleanup-recipient-headers (read-string "Recipients: "))))
 	  (t (error "mc-encrypt-generic: recipients not string or nil")))
     (or scheme (setq scheme mc-default-scheme))
-    (or start (setq start (point-min)))
-    (or end (setq end (point-max)))
     (if (funcall (cdr (assoc 'encryption-func (funcall scheme)))
 		 recipients start end from sign)
 	(progn
@@ -131,18 +132,12 @@ of `mc-default-scheme'.  Returns t on success, nil otherwise."
 	       (concat "^" (regexp-quote mail-header-separator) "$"))
 	      (narrow-to-region (point-min) (point))
 	      (and (featurep 'mailalias)
+		   (not (featurep 'mail-abbrevs))
 		   mail-aliases
 		   (expand-mail-aliases (point-min) (point-max)))
-	      (mapconcat
-	       'identity
-	       (append
-		(mc-cleanup-recipient-headers
-		 (or (mail-fetch-field "to" nil t) ""))
-		(mc-cleanup-recipient-headers
-		 (or (mail-fetch-field "bcc" nil t) ""))
-		(mc-cleanup-recipient-headers
-		 (or (mail-fetch-field "cc" nil t) "")))
-	       ", ")))
+	      (mc-strip-addresses
+	       (mapcar 'cdr
+		       (mc-get-fields "to\\|cc\\|bcc")))))
 
       (if (not from)
 	  (save-restriction
@@ -158,8 +153,6 @@ of `mc-default-scheme'.  Returns t on success, nil otherwise."
 		    default-recipients
 		  (read-from-minibuffer "Recipients: " default-recipients))))
      
-      (or recipients (error "No recipients!"))
-
       (or start (setq start headers-end))
       (or end (setq end (point-max-marker)))
 
@@ -224,7 +217,7 @@ inhibits clearsigning (pgp)."
   (let* ((mode-alist (cdr-safe (assq major-mode mc-modes-alist)))
 	 (func (or (cdr-safe (assq 'sign mode-alist))
 		   'mc-sign-generic))
-	 from scheme unclearsig)
+	 from scheme)
     (if (>= arg 16)
 	(setq scheme
 	      (cdr (assoc
@@ -237,12 +230,16 @@ inhibits clearsigning (pgp)."
 
 (defun mc-sign-generic (withkey scheme start end unclearsig)
   (or scheme (setq scheme mc-default-scheme))
-  (or start (setq start (point-min)))
-  (or end (setq end (point-max)))
+  (or start (setq start (point-min-marker)))
+  (or (markerp start) (setq start (copy-marker start)))
+  (or end (setq end (point-max-marker)))
+  (or (markerp end) (setq end (copy-marker end)))
   (run-hooks 'mc-pre-signature-hook)
-  (funcall (cdr (assoc 'signing-func (funcall scheme)))
-	   start end withkey unclearsig)
-  (run-hooks 'mc-post-signature-hook))
+  (if (funcall (cdr (assoc 'signing-func (funcall scheme)))
+	       start end withkey unclearsig)
+      (progn
+	(run-hooks 'mc-post-signature-hook)
+	t)))
 
 (defun mc-sign-message (&optional withkey scheme start end unclearsig)
   "Clear sign the message."
@@ -439,8 +436,7 @@ Exact behavior depends on current major mode."
 		 (if (cdr decryption-result)
 		     (rmail-add-label "verified")))
 		(t
-		 (let ((tmp (generate-new-buffer "*Mailcrypt Viewing*"))
-		       (map (make-sparse-keymap)))
+		 (let ((tmp (generate-new-buffer "*Mailcrypt Viewing*")))
 		   (copy-to-buffer tmp (point-min) (point-max))
 		   (rmail-abort-edit)
 		   (switch-to-buffer tmp t)
@@ -449,9 +445,8 @@ Exact behavior depends on current major mode."
 			   (current-time-string) "\n")
 		   (rmail-convert-file)
 		   (rmail-mode)
-		   (define-key map "q" 'mc-rmail-view-quit)
-		   (setq map (nconc map (current-local-map)))
-		   (use-local-map map)
+		   (use-local-map (copy-keymap (current-local-map)))
+		   (local-set-key "q" 'mc-rmail-view-quit)
 		   (set-buffer-modified-p nil)))))
       (if (eq major-mode 'rmail-edit-mode)
 	  (rmail-abort-edit)))))
@@ -473,15 +468,15 @@ Exact behavior depends on current major mode."
   (vm-select-folder-buffer)
   (vm-check-for-killed-summary)
   (vm-error-if-folder-empty)
-  (mc-verify-signature))
+  (save-restriction
+    (vm-widen-page)
+    (mc-verify-signature)))
 
 (defun mc-vm-decrypt-message ()
   "*Decrypt the contents of the current VM message"
   (interactive)
-  (let ((oldbuf (current-buffer))
-	decryption-result
-	from-line
-	condition-data)
+  (let ((vm-frame-per-edit nil)
+	from-line)
     (if (interactive-p)
 	(vm-follow-summary-cursor))
     (vm-select-folder-buffer)
@@ -503,7 +498,8 @@ Exact behavior depends on current major mode."
 	  ((and (not (eq mc-always-replace 'never))
 		(or mc-always-replace
 		    (y-or-n-p "Replace encrypted message with decrypted? ")))
-           (vm-edit-message-end))
+	   (let ((this-command 'vm-edit-message-end))
+	     (vm-edit-message-end)))
           (t
            (let ((tmp (generate-new-buffer "*Mailcrypt Viewing*")))
              (copy-to-buffer tmp (point-min) (point-max))
@@ -522,41 +518,66 @@ Exact behavior depends on current major mode."
   (vm-select-folder-buffer)
   (vm-check-for-killed-summary)
   (vm-error-if-folder-empty)
-  (mc-snarf-keys))
+  (save-restriction
+    (vm-widen-page)
+    (mc-snarf-keys)))
 
 ;;}}}
 ;;{{{ GNUS
 
-(defun mc-gnus-summary-verify-signature ()
+(defun mc-gnus-verify-signature ()
   (interactive)
-  (gnus-summary-select-article gnus-save-all-headers gnus-save-all-headers)
+  (gnus-summary-select-article)
   (gnus-eval-in-buffer-window gnus-article-buffer
     (save-restriction (widen) (mc-verify-signature))))
 
-(defun mc-gnus-summary-snarf-keys ()
+(defun mc-gnus-snarf-keys ()
   (interactive)
-  (gnus-summary-select-article gnus-save-all-headers gnus-save-all-headers)
+  (gnus-summary-select-article)
   (gnus-eval-in-buffer-window gnus-article-buffer
     (save-restriction (widen) (mc-snarf-keys))))
 
-(defun mc-gnus-summary-decrypt-message ()
+(defun mc-gnus-decrypt-message ()
   (interactive)
-  (gnus-summary-select-article gnus-save-all-headers gnus-save-all-headers)
-  (gnus-eval-in-buffer-window gnus-article-buffer
-    (save-restriction (widen) (mc-decrypt-message))))
+  (gnus-summary-select-article)
+  ;; Gnus 5 has the string "Gnus" instead of "GNUS" in gnus-version.
+  (if (not (let ((case-fold-search nil))
+	     (string-match "Gnus" gnus-version)))
+      (gnus-eval-in-buffer-window
+       gnus-article-buffer
+       (save-restriction (widen) (mc-decrypt-message)))
+    ;; Gnus 5 allows editing of articles.  (Actually, it makes a great
+    ;; mail reader.)
+    (gnus-eval-in-buffer-window gnus-article-buffer
+      (gnus-summary-edit-article t)
+      (save-restriction
+	(widen)
+	(cond ((not (car (mc-decrypt-message)))
+	       (gnus-summary-edit-article-postpone))
+	      ((and (not (gnus-group-read-only-p))
+		    (not (eq mc-always-replace 'never))
+		    (or mc-always-replace
+			(y-or-n-p
+			 "Replace encrypted message on disk? ")))
+	       (gnus-summary-edit-article-done))
+	      (t
+	       (gnus-summary-edit-article-postpone)))))))
 
 ;;}}}		
 ;;{{{ MH
 
-(defun mc-mh-decrypt-message (&optional decrypt-on-disk)
-  "*Decrypt the contents of the current MH message in the show buffer.
-With prefix arg, decrypt the message on disk as well."
+(defun mc-mh-decrypt-message ()
+  "Decrypt the contents of the current MH message in the show buffer."
   (interactive "P")
   (let* ((msg (mh-get-msg-num t))
 	 (msg-filename (mh-msg-filename msg))
 	 (show-buffer (get-buffer mh-show-buffer))
-	 decrypt-okay)
-    (setq decrypt-on-disk (or mc-always-replace decrypt-on-disk))
+	 decrypt-okay decrypt-on-disk)
+    (setq
+     decrypt-on-disk
+     (and (not (eq mc-always-replace 'never))
+	  (or mc-always-replace
+	      (y-or-n-p "Replace encrypted message on disk? "))))
     (if decrypt-on-disk
 	(progn
 	  (save-excursion
@@ -590,20 +611,17 @@ With prefix arg, decrypt the message on disk as well."
 (defun mc-mh-verify-signature ()
   "*Verify the signature in the current MH message."
   (interactive)
-  (let ((msg (mh-get-msg-num t)))
-    (mh-show msg)
-    (save-excursion
-      (set-buffer mh-show-buffer)
-      (mc-verify-signature))))
+  (mh-show)
+  (mh-in-show-buffer (mh-show-buffer)
+    (mc-verify-signature)))
+    
 
 (defun mc-mh-snarf-keys ()
   (interactive)
-  (mh-show (mh-get-msg-num t))
-  (save-excursion
-    (set-buffer mh-show-buffer)
+  (mh-show)
+  (mh-in-show-buffer (mh-show-buffer)
     (mc-snarf-keys)))
 
-
 ;;}}}
 
 ;;}}}
